/*
******************************************************************************

   LibBUF: plain functions

   Copyright (C) 2001 Christoph Egger   [Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <ggi/internal/buf.h>
#include <ggi/internal/buf_debug.h>
#include <ggi/internal/galloc_ops.h>

#include <stdarg.h>

/* Checks to see if a GA_RT_RENDERER_DRAWOPS motor is already
 * in the list, and returns it if found. 
 */
static ggiGA_resource_handle _ggiBufFindRendererMotor(ggiGA_resource_list reslist) {
	ggiGA_resource_handle res;

	res = reslist;
	while (res != NULL) {
		if (ggiGAIsMotor(res) && !ggiGAIsCap(res) &&
		    (ggiGAGetType(res) == GA_RT_RENDERER_DRAWOPS))
			break;
		res = ggiGAGetNext(res);		
	};
	return res;
}

/* Checks to see if a GA_RT_RENDERER_DRAWOPS carb is already
 * in the compound, and returns it if found. 
 */
static ggiGA_resource_handle _ggiBufFindRendererCarb(ggiGA_resource_handle head) {
	ggiGA_resource_handle res, last;

	last = ggiGAFindLastSeeabove(NULL, head);
	res = head;
	while (res != NULL) {
		if (ggiGAIsCarb(res) && !ggiGAIsCap(res) &&
		    (ggiGAGetType(res) == GA_RT_RENDERER_DRAWOPS))
			break;
		if (res == last) return NULL;
		res = ggiGAGetNext(res);		
	};
	return res;
}

int ggiBufAdd(ggi_visual_t vis, 
	      ggiGA_resource_list *reqlist,
	      enum ggiGA_resource_type fulltype,
	      struct ggiGA_resource_props *props,
	      enum ggiBuf_flags flags,
	      ggiGA_resource_handle tothis,
	      ggiGA_resource_handle *handle)

{
	int rc = GALLOC_OK;
	ggiGA_resource_handle motor_h;
	struct ggiGA_resource_props carb_p;
	ggiGA_resource_handle tmp, tmp2;

	BUFDPRINT_CORE("ggiBufAdd(%p,%p,%x,%p,%x,%p,%p) called\n",
		       vis, reqlist, fulltype, props, flags, tothis, handle);

	if (reqlist == NULL) return GGI_EARGINVAL;
	if (props == NULL) return GGI_EARGINVAL;
	if (handle == NULL) return GGI_EARGINVAL;

	if (tothis != NULL) {
		tothis = ggiGAFindCompoundHead(*reqlist, tothis);
		if (tothis == NULL) return GGI_EARGINVAL;
		/* TODO: support for "floaters" */
		if (ggiGAGetType(tothis) != GA_RT_FRAME) goto notvis;
	}	/* if */
	else {
		/* TODO: Find the last mode even if it is not in service. */
		tothis = ggiGAFindLastMode(*reqlist, NULL);
	}
	if (tothis == NULL) {
	  fprintf(stderr, "Nothing to attach to (floaters unimplemented.)\n");
	  return GALLOC_EFAILED;
	}

	motor_h = _ggiBufFindRendererMotor(*reqlist);
	if (motor_h == NULL) {
		/* The renderer motor has not been added, so we must add it. */
		struct ggiGA_resource_props renderer;

		BUFDPRINT("Adding GA_RT_RENDERER_DRAWOPS motor\n");

		ggiBufClearMotorProperties(&renderer);
		rc = ggiGAAdd(reqlist, &renderer, 
			      GA_RT_MOTOR | GA_RT_RENDERER_DRAWOPS, &motor_h);
		if (rc != GALLOC_OK) return GALLOC_EFAILED;
		BUF_ASSERT(motor_h != NULL, "Failure in basic GAlloc call");
	}	/* if */

	ggiBufClearCarbProperties(&carb_p);
	switch (fulltype) {
	case GA_RT_BUFFER_ABUFFER:
		if (flags & BUF2GA_FLAGS_INVALID4ALPHA_VIS) {
			BUFDPRINT("Bad flags (%x) for Alpha buffer\n", flags);
			return GALLOC_EFAILED;
		}	/* if */
		carb_p.sub.carb.must.rop  = 0;
		carb_p.sub.carb.can.rop   = LL_ROP_ALPHA_AUX;
		carb_p.sub.carb.must.math = 0;
		carb_p.sub.carb.can.math  = flags;
		carb_p.sub.carb.can.math  &= BUF2GA_ALPHA_BLEND_MASK |
		  (BUF2GA_ALPHA_BLEND_MASK << BUF_ALTBLEND_SHIFT);
		break;
	case GA_RT_BUFFER_ZBUFFER:
		if (flags & BUF2GA_FLAGS_INVALID4Z) {
			BUFDPRINT("Bad flags (%x) for Z buffer\n", flags);
			return GALLOC_EFAILED;
		}	/* if */
		carb_p.sub.carb.must.rop  = 0;
		carb_p.sub.carb.can.rop   = LL_ROP_Z;
		carb_p.sub.carb.must.math = 0;
		carb_p.sub.carb.can.math = flags & BUF2GA_Z_MASK;
		if (flags & BUF_DISPOSE) 
		  carb_p.sub.carb.can.math |= LL_MATH_Z_DISPOSE;
		if (flags & (BUF_DISPOSE << BUF_ALTBLEND_SHIFT))
		  carb_p.sub.carb.can.math |= 
		    LL_MATH_Z_DISPOSE << LL_MATH_ALTSHIFT;
	  break;
	default:
		BUFDPRINT("Don't know how to handle type %x yet\n", fulltype);
		return GALLOC_ESPECIFY;
	}	/* switch */

	tmp = ggiGAFindLastSeeabove(*reqlist, tothis);
	if (ggiGAIsCap(ggiGAGetNext(tmp))) tmp = ggiGAGetNext(tmp);

	/* If there are either A) resources below our target compound, 
	 * or B) a target carb which is in the NORESET state, we cannot simply 
	 * append, we have to generate another copy of the compound
	 * and attach it to the end of the list. 
	 */
	tmp2 = _ggiBufFindRendererCarb(tothis);
	if ((ggiGAGetNext(tmp) != NULL) || 
	    ((tmp2 != NULL) && (ggiGAGetState(tmp2) & GA_STATE_NORESET))) {

		/* Make sure all tanks have a tag group, else the append
		 * operation will not go as planned.
		 */
		tmp2 = tothis;
		do {
			if (ggiGAIsTank(tmp2) && !ggiGAIsCap(tmp2))
				ggiGASetAutoTag(*reqlist, tmp2);
			if (tmp == tothis) break;
			tmp2 = ggiGAGetNext(tmp2);
		} while (tmp2 != tmp);
		rc = ggiGAAddMutualCompound(*reqlist, &tothis);
		if (rc != GALLOC_OK) {
			fprintf(stderr, "ggiGAAddMutualCompound failed.\n");
			return rc;
		}
		/* TODO: take NOCHANGE and NORESET flags off new carb */
	}	/* if */
	tmp = _ggiBufFindRendererCarb(tothis);
	if (tmp != NULL) {
		ggiGAGetProperties(tmp)->sub.carb.can.math |= 
		  carb_p.sub.carb.can.math;
		ggiGAGetProperties(tmp)->sub.carb.can.rop  |= 
		  carb_p.sub.carb.can.rop;
	} else {
		rc = ggiGAAddSeeabove(reqlist, &carb_p, GA_RT_CARB | 
				      GA_RT_RENDERER_DRAWOPS, &tmp);
		rc = ggiGATagOnto(*reqlist, motor_h, tmp);
	}	/* if */
	rc = ggiGAAddSeeabove(reqlist, props, fulltype, handle);

	return rc;

 notvis:
	/* TODO: non-frame resources */
	return GALLOC_ESPECIFY;

}	/* ggiBufAdd */



ggiBuf_t ggiBufCreateBuffer(ggi_visual_t vis, 
			    enum ggiGA_resource_type fulltype,
			    struct ggiGA_resource_props *props,
			    enum ggiBuf_flags flags, ggiBuf_t forthis)
{

	ggiBuf_t buf;
	ggiGA_resource_list reqlist = NULL;
	ggiGA_resource_list reslist = NULL;
	ggiGA_resource_handle req_h, res_h;
	int rc;

	BUFDPRINT_CORE("ggiBufCreateBuffer(%p,%x,%p,%x,%p) called\n",
		       vis, fulltype, props, flags, forthis);

	/* Prime the pump -- in case no mode resource in visual yet. */
	if (*(LIBGGI_BUFEXT(vis)->reqlist) == NULL) {
		BUFDPRINT_CORE("Jump starting LibGAlloc (reqlist %p)\n", 
			       *(LIBGGI_BUFEXT(vis)->reqlist));
		ggiGAGet(vis, LIBGGI_BUFEXT(vis)->reqlist);
	}	/* if */

	/* Copy the current resource list into the local reqlist,
	 * and add our buffer to it.
	 */
	rc = ggiGAGet(vis, &reqlist);
	LIB_ASSERT(rc == GALLOC_OK);
	if (rc != GALLOC_OK) return NULL;

	if (forthis == NULL) {
		rc = ggiBufAdd(vis, &reqlist, fulltype, props, flags, 
			       NULL, &req_h);
	} else {
		/* TODO: Add buffer to object other than visual. */
#if 0
		if (forthis->type == 0) goto err0; /* Not a real buffer */
		rc = ggiBufAdd(vis, &reqlist, fulltype, props, flags,
			       forthis->parts.res.tank, &req_h);
#endif
	}	/* if */

	LIB_ASSERT(rc == GALLOC_OK);
	if (rc != GALLOC_OK) goto err0;

	BUFDPRINT_CORE("Checking resources.\n");
	rc = ggiGACheck(vis, reqlist, NULL);
	BUFDPRINT("%s:%s:%i: ggiGACheck() returns: %i\n", DEBUG_INFO, rc);
	if (rc != GALLOC_OK) goto err0;

	BUFDPRINT_CORE("Allocating resources.\n");
	rc = ggiGASet(vis, reqlist, &reslist);
	BUFDPRINT("%s:%s:%i: ggiGASet() returns: %i\n", DEBUG_INFO, rc);
	if (rc != GALLOC_OK) goto err1;

	/* Get the handle of the filled-out resource. */
        res_h = ggiGAHandle(reslist, reqlist, req_h);
        LIB_ASSERT(res_h != NULL);
        if (res_h == NULL) goto err1;
	
	LIB_ASSERT(ggiGAFindCompoundHead(reslist, res_h) != NULL);
	LIB_ASSERT(_ggiBufFindRendererCarb(ggiGAFindCompoundHead(reslist, res_h)) != NULL);
	LIB_ASSERT(_ggiBufFindRendererMotor(reslist) != NULL);

	buf = LIBGGI_BUFEXT(vis)->_tie(vis, _ggiBufFindRendererMotor(reslist),
				       _ggiBufFindRendererCarb(ggiGAFindCompoundHead(reslist, res_h)),
				       res_h,
				       0, NULL);
	if (buf == NULL) goto err2;

	rc = ggiGAEmptyList(&reqlist);
	LIB_ASSERT(rc == GALLOC_OK);

	/* perform touch-up surgery on the reqlist */
	rc = ggiGASpliceList(LIBGGI_BUFEXT(vis)->reqlist, &reslist);
	LIB_ASSERT(rc == GALLOC_OK);

	BUFDPRINT("%s:%s:%i: completed successfully\n", DEBUG_INFO);
	return buf;

 err2:
	BUFDPRINT("%s:%s:%i: err2\n", DEBUG_INFO);
	rc = ggiGARelease(vis, &reslist, &res_h);
	LIB_ASSERT(rc == GALLOC_OK);
 err1:
	BUFDPRINT("%s:%s:%i: err1\n", DEBUG_INFO);
	rc = ggiGAEmptyList(&reslist);
	LIB_ASSERT(rc == GALLOC_OK);
 err0:
	BUFDPRINT("%s:%s:%i: err0\n", DEBUG_INFO);
	rc = ggiGAEmptyList(&reqlist);
	LIB_ASSERT(rc == GALLOC_OK);

	BUFDPRINT("%s:%s:%i: returning with error %i\n", DEBUG_INFO, rc);
	return NULL;

}	/* ggiBufCreateBuffer */




ggiBuf_t ggiBufCreateZ(ggi_visual_t vis, ggiBuf_t forthis,
			enum ggiGA_storage_type stype,
			ggi_graphtype gt, enum ggiBuf_flags flags)
{
	struct ggiGA_resource_props props;

	BUFDPRINT_CORE("ggiBufCreateZ(%p,%p,%x,%x,%x) called\n",
		       vis, forthis, stype, gt, flags);

	ggiBufClearTankProperties(&props);

	props.sub.tank.gt = gt;
	props.storage_need = props.storage_ok = stype;

	return ggiBufCreateBuffer(vis, GA_RT_BUFFER_ZBUFFER,
				  &props, flags, forthis);

}	/* ggiBufCreateZ */


/* Same for alpha */
ggiBuf_t ggiBufCreateAlpha(ggi_visual_t vis, ggiBuf_t forthis,
			enum ggiGA_storage_type stype,
			ggi_graphtype gt, enum ggiBuf_flags flags)
{
	struct ggiGA_resource_props props;

	BUFDPRINT_CORE("ggiBufCreateAlpha(%p,%p,%x,%x,%x) called\n",
		       vis, forthis, stype, gt, flags);

	ggiBufClearTankProperties(&props);

	props.sub.tank.gt = gt;
	props.storage_need = props.storage_ok = stype;

	return ggiBufCreateBuffer(vis, GA_RT_BUFFER_ABUFFER,
				  &props, flags, forthis);

}	/* ggiBufCreateAlpha */


int ggiBufDestroyBuffer(ggiBuf_t *buf)
{
	ggiGA_resource_handle handle = NULL;
	ggi_visual_t vis;
	int rc;
#if 0
	if (buf == NULL) return GGI_EARGINVAL;
	if (buf[0] == NULL) return GGI_EARGINVAL;

	LIB_ASSERT(buf[0]->vis == ggiGAGetVisual(buf[0]->res_handle));
	vis = buf[0]->vis;

	if (buf[0]->type == 0) {
		/* TODO: Not a real buffer */
		free(buf[0]);
		buf[0] = NULL;
		return GALLOC_OK;
	}

	handle = buf[0]->req_handle;
	LIB_ASSERT(handle != NULL);

	rc = ggiBufUntie(buf);
	LIB_ASSERT(rc == GALLOC_OK);

#warning check if other ggiBuf_t are using this resource.
	rc = ggiGARelease(vis, LIBGGI_BUFEXT(vis)->reqlist, &handle);
	LIB_ASSERT(rc == GALLOC_OK);

	return rc;
#endif
	return 0;

}	/* ggiBufDestroyBuffer */


ggiGA_resource_handle ggiBufGetMotorHandle(ggiBuf_t buf)
{
	ggiGA_resource_handle *buf_ga;

	LIB_ASSERT(buf != NULL);
	if (buf == NULL) return NULL;

	buf_ga = GA_LL(buf);
	LIB_ASSERT(buf_ga != NULL);
        if (buf_ga == NULL) return NULL;

        while (*buf_ga) {
          if (ggiGAIsMotor(*buf_ga)) return(*buf_ga);
          buf_ga++;
        }

	return NULL;
}	/* ggiBufGetMotorHandle */

ggiGA_resource_handle ggiBufGetCarbHandle(ggiBuf_t buf)
{
	ggiGA_resource_handle *buf_ga;

	LIB_ASSERT(buf != NULL);
	if (buf == NULL) return NULL;

	buf_ga = GA_LL(buf);
	LIB_ASSERT(buf_ga != NULL);
        if (buf_ga == NULL) return NULL;

        while (*buf_ga) {
          if (ggiGAIsCarb(*buf_ga)) return(*buf_ga);
          buf_ga++;
        }

	return NULL;

}	/* ggiBufGetCarbHandle */

ggiGA_resource_handle ggiBufGetTankHandle(ggiBuf_t buf)
{
	ggiGA_resource_handle *buf_ga;

	LIB_ASSERT(buf != NULL);
	if (buf == NULL) return NULL;

	buf_ga = GA_LL(buf);
	LIB_ASSERT(buf_ga != NULL);
        if (buf_ga == NULL) return NULL;

        while (*buf_ga) {
          if (ggiGAIsTank(*buf_ga)) return(*buf_ga);
          buf_ga++;
        }

	return NULL;
}	/* ggiBufGetTankHandle */


/* Fills properties with useful default values */
int ggiBufClearMotorProperties(struct ggiGA_resource_props *prop)
{
	int rc = GALLOC_OK;
	LIB_ASSERT(prop != NULL);

	rc = ggiGAClearMotorProperties(prop);
	LIB_ASSERT(rc == GALLOC_OK);

	return rc;
}	/* ggiBufClearMotorProperties */


int ggiBufClearCarbProperties(struct ggiGA_resource_props *prop)
{
	int rc = GALLOC_OK;
	LIB_ASSERT(prop != NULL);

	rc = ggiGAClearCarbProperties(prop);
	LIB_ASSERT(rc == GALLOC_OK);

	return rc;
}	/* ggiBufClearCarbProperties */


int ggiBufClearTankProperties(struct ggiGA_resource_props *prop)
{
	int rc = GALLOC_OK;
	LIB_ASSERT(prop != NULL);

	rc = ggiGAClearTankProperties(prop);
	LIB_ASSERT(rc == GALLOC_OK);

	return rc;
}	/* ggiBufTankProperties */

